require 'sinatra/base'
require 'lib/sinatra_helpers'
require 'lib/console_auth'
require 'yaml'
require 'active_record'
require 'mail'

class ConsoleAuthApp < Sinatra::Base

  helpers SinatraHelpers

  configure :development, :test do
    enable :raise_errors
  end

  get '/admin' do
    erb :admin, {:layout => false}
  end

  get %r[^(/views/.*)] do
    send_file "batman" + params[:captures].first
  end

  get '/activate/:key' do
    cache_control :no_cache
    @min_length = ConsoleAuth.get_config['password']['min_length']
    erb :activate
  end

  get '/account' do
    cache_control :no_cache
    @min_length = ConsoleAuth.get_config['password']['min_length']
    @user = ConsoleAuth.get_user(env)
    return [ 400, "No user was found."] unless @user
    erb :account
  end

  get '/account_types' do
    account_types = []
    ConsoleAuth.authorization_defaults.each do |key, type|
      account_types.push({'id' => key, 'type' => type['description'], 'role' => type['default_role'] })
    end
    { 'account_types' => account_types }.to_json
  end

  get '/users' do
    cache_control :no_cache
    { 'users' => user_info(AuthorizedUser.all, env, ConsoleAuth.get_user(env)) }.to_json
  end

  get '/users/:id' do
    cache_control :no_cache
    user = AuthorizedUser.find_by_id(params[:id])
    return [ 404, "No user with id #{params[:id]} was found" ] unless user
    user_info([user]).first.to_json
  end

  post '/users' do
    cache_control :no_cache
    content_type :json
    begin
      admin = ConsoleAuth.get_user(env)
      return [ 400, "You must be an admin to do that!" ] unless admin.roles.first.name == "admin"
      role = Role.find_by_name(params[:user][:role])
      return [ 404, "No role with name #{params[:user][:role]} was found" ] unless role
      unless params[:user][:role] && params[:user][:username]
        return [ 400, "Required information for creating a user (email, role) was not passed." ]
      end
      auth_user = AuthorizedUser.new({ :username => params[:user][:username],
                        :roles => [role],
                        :origin => "local",
                        :status => "pending" })
      auth_user.save!
      user = User.new({ :username => params[:user][:username]})
      user.password = ''
      user.save!
      ConsoleAuth.logger.info("User #{user.id} was created by admin #{admin.id} (#{admin.username}).")
      activation = create_activation(auth_user, admin.username)
      send_confirmation_email(auth_user, admin.username, activation.created_at, activation_link(activation.key, env['rack.url_scheme'], env['SERVER_PORT']))
    rescue Exception => e
      ConsoleAuth.logger.error("An error occured while creating user #{params[:user][:username]}: #{e.message}")
      ConsoleAuth.logger.error("Stack trace: #{e.backtrace}")
      error_message = e.message =~ /has already been taken/ ? "User already exists" : e.message
      return [ 400, error_message ]
    end
    info = user_info([auth_user]).first
    info['activation_link'] = activation_link(activation.key, env['rack.url_scheme'], env['SERVER_PORT'])
    info.to_json
  end

  delete '/users/:id' do
    cache_control :no_cache
    begin
      admin = ConsoleAuth.get_user(env)
      return [ 400, "You must be an admin to do that!" ] unless admin.roles.first.name == "admin"
      auth_user = AuthorizedUser.find_by_id(params[:id])
      return [ 404, "No user with id #{params[:id]} was found" ] unless auth_user
      return [ 400, "Admins cannot delete themselves" ] if admin.id == auth_user.id
      if auth_user.status == "pending" || auth_user.status == "expired"
        activation = Activation.find_by_user_id(auth_user.id)
        activation.delete unless activation.nil?
      end
      if auth_user.origin = "local" then
        user = User.find_by_username(auth_user.username)
        user.delete unless user.nil?
      end
      auth_user.delete unless auth_user.nil?
      ConsoleAuth.logger.info("User #{params[:id]} was deleted by admin #{admin.id} (#{admin.username}).")
    rescue Exception => e
      ConsoleAuth.logger.error("An error occured while deleting user #{params[:id]} (#{params[:username]}): #{e.message}")
      return [ 400, e.message ]
    end
    "User was deleted.".to_json
  end

  put '/users/:id' do
    cache_control :no_cache
    begin
      activation = nil
      admin = ConsoleAuth.get_user(env)
      return [ 400, "You must be an admin to do that!" ] unless admin.roles.first.name == "admin"
      auth_user = AuthorizedUser.find_by_id(params[:id])
      return [ 404, "No user with id #{params[:id]} was found" ] unless auth_user
      if admin.id == auth_user.id && (params[:user][:status] != auth_user.status)
        return [ 400, "Admins cannot disable themselves" ]
      end

      if auth_user.origin == "local" then
        user = User.find_by_username(auth_user.username)
        user.attributes.each do |attr, val|
          param_val = params[:user][attr.to_sym]
          next unless param_val && val.to_s != param_val.to_s
          user.update_attribute(attr, param_val)
          ConsoleAuth.logger.info("User #{user.id} (#{user.username}) was updated by admin #{admin.id} (#{admin.username}): #{attr} changed from #{val} to #{param_val}")
        end
      end

      auth_user.attributes.each do |attr, val|
        param_val = params[:user][attr.to_sym]
        next unless param_val && val.to_s != param_val.to_s
        auth_user.update_attribute(attr, param_val) unless attr == 'origin'
        if attr == "username"
          activation = create_activation(auth_user, admin.username)
          send_confirmation_email(auth_user, admin.username, activation.created_at, activation_link(activation.key, env['rack.url_scheme'], env['SERVER_PORT']))
          auth_user.update_attribute("status","pending")
        end
        ConsoleAuth.logger.info("AuthorizedUser #{auth_user.id} (#{auth_user.username}) was updated by admin #{admin.id} (#{admin.username}): #{attr} changed from #{val} to #{param_val}")
      end

      old_role = auth_user.roles.first.name
      if params[:user][:role] && params[:user][:role] != old_role
        return [ 400, "Admins cannot change their own role" ] if admin.id == auth_user.id
        auth_user.roles = [Role.find_by_name(params[:user][:role])]
        auth_user.save!
        ConsoleAuth.logger.info("AuthorizedUser #{auth_user.id} (#{auth_user.username}) was updated by admin #{admin.id} (#{admin.username}): role changed from #{old_role} to #{auth_user.roles.first.name}")
      end
    rescue Exception => e
      ConsoleAuth.logger.error("An error occured while attempting to update user #{auth_user.id} (#{auth_user.username}): #{e.inspect}")
      ConsoleAuth.logger.error("Stacktrace: #{e.backtrace}")
      return [ 400, e.message ]
    end
    info = user_info([auth_user]).first
    info['activation_link'] = activation_link(activation.key, env['rack.url_scheme'], env['SERVER_PORT']) if activation
    info.to_json
  end

  post '/unset_password' do
    cache_control :no_cache
    content_type :json
    begin
      current_user = ConsoleAuth.get_user(env)
      return [ 400, "You must be an admin to do that!" ] unless current_user.roles.first.name == "admin"
      return [ 400, "Missing parameter: user_id"] unless params[:user_id]
      auth_user = AuthorizedUser.find_by_id(params[:user_id])
      return [ 400, "User is externally managed."] unless auth_user.origin == "local"
      target_user = User.find_by_username(auth_user.username)
      target_user.password = ''
      auth_user.status = "pending"
      auth_user.save!
      target_user.save!
      ConsoleAuth.logger.info("User #{target_user.id} (#{target_user.username}) had their password reset by admin #{current_user.id} (#{current_user.username})")
      activation = create_activation(auth_user, current_user.username)
      send_confirmation_email(auth_user, current_user.username, activation.created_at, activation_link(activation.key, env['rack.url_scheme'], env['SERVER_PORT']))
      result = user_info([auth_user]).first
      result['activation_link'] = activation_link(activation.key, env['rack.url_scheme'], env['SERVER_PORT'])
      return result.to_json
    rescue Exception => e
      ConsoleAuth.logger.error("An error occured when #{current_user.id} (#{current_user.username}) attempted to unset the password for user #{target_user.id} (#{target_user.username}): #{e.inspect}")
      ConsoleAuth.logger.error("Stacktrace: #{e.backtrace}")
      return [ 400, e.message ]
    end
  end

  post '/reset_password' do
    cache_control :no_cache
    content_type :json
    begin
      auth_user = ConsoleAuth.get_user(env)
      return [ 400, "User is externally managed."] unless auth_user.origin == "local"
      current_user = User.find_by_username(auth_user.username)
      new_password = params[:password]
      return [ 400, "Missing parameter: password"] unless new_password
      min_length = ConsoleAuth.get_config['password']['min_length']
      return [ 400, "Password must be #{min_length} characters or longer." ] if new_password.length < min_length
      current_user.password = new_password
      current_user.save!
      ConsoleAuth.logger.info("User #{current_user.id} (#{current_user.username}) reset their password.")
      return "Password successfully reset.".to_json
    rescue Exception => e
      ConsoleAuth.logger.error("An error occured when #{current_user.id} (#{current_user.username}) attempted to reset their password: #{e.inspect}")
      ConsoleAuth.logger.error("Stacktrace: #{e.backtrace}")
      return [ 400, e.message ]
    end
  end

  get '/activation/:key' do
    cache_control :no_cache
    content_type :json
    activation = Activation.find_by_key(params[:key])
    return [ 404, "Invalid activation link"] unless activation
    auth_user = AuthorizedUser.find_by_id(activation.user_id)
    user = User.find_by_username(auth_user.username)
    return [ 400, "No user with id #{activation.user_id} was found corresponding to activation record #{activation.id}"] unless user
    if (auth_user.status == "pending") && fresh?(activation)
      return user_info([auth_user]).first.to_json
    else
      auth_user.status = "expired"
      auth_user.save!
      ConsoleAuth.logger.info("User #{user.id} (#{user.username}) attempted to activate an expired account; delete the account and create a new one.")
      return [ 400, "Activation has expired; notify your admin to create a new account for you." ]
    end
  end

  post '/activation/:key' do
    cache_control :no_cache
    begin
      activation = Activation.find_by_key(params[:key])
      return [ 404, "Invalid activation link"] unless activation
      auth_user = AuthorizedUser.find_by_id(activation.user_id)
      user = User.find_by_username(auth_user.username)
      return [ 404, "No user with id #{activation.user_id} was found corresponding to activation record #{activation.id}"] unless user
      password = params[:password]
      return [ 400, "Missing parameter: password"] unless password
      min_length = ConsoleAuth.get_config['password']['min_length']
      return [ 400, "Password must be #{min_length} characters or longer." ] if password.length < min_length
      if (auth_user.status == "pending") && fresh?(activation)
        auth_user.status = "enabled"
        user.password = password
        user.save!
        auth_user.save!
        activation.delete
        ConsoleAuth.logger.info("User #{auth_user.id} (#{auth_user.username}) was activated")
      else
        auth_user.status = "expired"
        auth_user.save!
        ConsoleAuth.logger.info("User #{auth_user.id} (#{auth_user.username}) attempted to activate an expired account; delete the account and create a new one.")
        return [ 400, "Activation has expired; notify your admin to create a new account for you." ]
      end
    rescue Exception => e
      ConsoleAuth.logger.error("Activating user with activation key #{params[:key]} failed: #{e.inspect}")
      return [ 400, "Activating the user failed: #{e.message}" ]
    end
    "User has been activated.".to_json
  end
end
